/// <reference types="node" />
import { Account } from '@solana/spl-token';
import { Commitment, Connection, PublicKey, PublicKeyInitData, Transaction } from '@solana/web3.js';
import { BigNumber } from 'ethers';
import { TokenId, ChainName, ChainId, NATIVE, ParsedMessage, Context, ParsedRelayerPayload, Wormhole, TokenBridgeAbstract, SolanaAbstract } from '@wormhole-foundation/connect-sdk';
import { SolContracts } from './contracts';
/**
 * @category Solana
 */
export declare class SolanaContext extends TokenBridgeAbstract<Transaction> implements SolanaAbstract {
    readonly type = Context.SOLANA;
    readonly contracts: SolContracts;
    protected wormhole: Wormhole;
    connection: Connection | undefined;
    constructor(wormholeInstance: Wormhole);
    /**
     * Sets the Connection
     *
     * @param connection The Solana Connection
     */
    setConnection(connection: Connection): Promise<void>;
    fetchTokenDecimals(tokenAddr: string, chain?: ChainName | ChainId): Promise<number>;
    /**
     * Gets the owner address of an Associated Token Account
     *
     * @param accountAddr The associated token account address
     * @returns The owner address
     */
    getTokenAccountOwner(accountAddr: string): Promise<string>;
    getNativeBalance(walletAddress: string, chain: ChainName | ChainId): Promise<BigNumber>;
    getTokenBalance(walletAddress: string, tokenId: TokenId, chain: ChainName | ChainId): Promise<BigNumber | null>;
    /**
     * Gets the Associate Token Address
     *
     * @param token The token id (home chain/address)
     * @param account The wallet address
     * @returns The associated token address
     */
    getAssociatedTokenAddress(token: TokenId, account: PublicKeyInitData): Promise<PublicKey>;
    /**
     * Gets the Associate Token Account
     *
     * @param token The token id (home chain/address)
     * @param account The wallet address
     * @returns The account, or null if it does not exist
     */
    getAssociatedTokenAccount(token: TokenId, account: PublicKeyInitData): Promise<Account | null>;
    /**
     * Creates the Associated Token Account for a given wallet address. This must exist before a user can send a token bridge transfer, also it is the recipient address when sending the transfer.
     *
     * @param token The token id (home chain/address)
     * @param account The wallet address
     * @param commitment The commitment level
     * @returns The transaction for creating the Associated Token Account
     */
    createAssociatedTokenAccount(token: TokenId, account: PublicKeyInitData, commitment?: Commitment): Promise<Transaction | void>;
    /**
     * Prepare the transfer instructions for a native token bridge transfer from Solana
     *
     * @dev This _must_ be claimed on the destination chain, see {@link Wormhole#completeTransfer | completeTransfer}
     *
     * @param senderAddress The address of the sender
     * @param amount The token amount to be sent
     * @param recipientChain The destination chain name or id
     * @param recipientAddress The associated token address where funds will be sent
     * @param relayerFee The fee that would be paid to a relayer
     * @param payload Arbitrary bytes that can contain any addition information about a given transfer
     * @param commitment The commitment level
     * @returns The transaction for sending Native SOL from Solana
     */
    private transferNativeSol;
    /**
     * Prepare the transfer instructions for a token bridge transfer from Solana
     *
     * @dev This _must_ be claimed on the destination chain, see {@link Wormhole#completeTransfer | completeTransfer}
     *
     * @param senderAddress The address of the sender
     * @param amount The token amount to be sent
     * @param recipientChain The destination chain name or id
     * @param recipientAddress The associated token address where funds will be sent
     * @param fromAddress The token account pubkey, owned by fromOwner address
     * @param tokenChainId The id of the token's chain
     * @param mintAddress The token address on the destination
     * @param fromOwnerAddress If not specified, will default to the sender address
     * @param relayerFee The fee that would be paid to a relayer
     * @param payload Arbitrary bytes that can contain any addition information about a given transfer
     * @param commitment The commitment level
     * @returns The transaction for sending tokens from Solana
     */
    private transferFromSolana;
    startTransfer(token: TokenId | typeof NATIVE, amount: bigint, sendingChain: ChainName | ChainId, senderAddress: string, recipientChain: ChainName | ChainId, recipientAddress: string, relayerFee?: string, commitment?: Commitment): Promise<Transaction>;
    startTransferWithPayload(token: TokenId | typeof NATIVE, amount: bigint, sendingChain: ChainName | ChainId, senderAddress: string, recipientChain: ChainName | ChainId, recipientAddress: string, payload: Uint8Array | Buffer, commitment?: Commitment): Promise<Transaction>;
    formatAddress(address: PublicKeyInitData): Uint8Array;
    parseAddress(address: string): string;
    formatAssetAddress(address: string): Promise<Uint8Array>;
    parseAssetAddress(address: string): Promise<string>;
    getForeignAsset(tokenId: TokenId, chain: ChainName | ChainId): Promise<string | null>;
    mustGetForeignAsset(tokenId: TokenId, chain: ChainName | ChainId): Promise<string>;
    parseMessageFromTx(tx: string, chain: ChainName | ChainId): Promise<ParsedMessage[]>;
    completeTransfer(destChain: ChainName | ChainId, signedVAA: Uint8Array, overrides: any, payerAddr?: PublicKeyInitData): Promise<Transaction>;
    isTransferCompleted(destChain: ChainName | ChainId, signedVaa: string): Promise<boolean>;
    getCurrentBlock(): Promise<number>;
    parseRelayerPayload(payload: Buffer): ParsedRelayerPayload;
}
