"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SolanaContext = void 0;
const spl_token_1 = require("@solana/spl-token");
const web3_js_1 = require("@solana/web3.js");
const ethers_1 = require("ethers");
const utils_1 = require("ethers/lib/utils");
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const contracts_1 = require("./contracts");
const tokenBridge_1 = require("./utils/tokenBridge");
const wormhole_1 = require("./utils/wormhole");
const utils_2 = require("./utils");
const SOLANA_SEQ_LOG = 'Program log: Sequence: ';
const SOLANA_CHAIN_NAME = connect_sdk_1.CONFIG.MAINNET.chains.solana.key;
const SOLANA_MAINNET_EMMITER_ID = 'ec7372995d5cc8732397fb0ad35c0121e0eaa90d26f828a534cab54391b3a4f5';
const SOLANA_TESTNET_EMITTER_ID = '3b26409f8aaded3f5ddca184695aa6a0fa829b0c85caf84856324896d214ca98';
/**
 * @category Solana
 */
class SolanaContext extends connect_sdk_1.TokenBridgeAbstract {
    constructor(wormholeInstance) {
        super();
        this.type = connect_sdk_1.Context.SOLANA;
        this.wormhole = wormholeInstance;
        const tag = this.wormhole.network === connect_sdk_1.Network.MAINNET ? 'mainnet-beta' : 'devnet';
        this.connection = new web3_js_1.Connection(this.wormhole.conf.rpcs.solana || (0, web3_js_1.clusterApiUrl)(tag));
        this.contracts = new contracts_1.SolContracts(this.wormhole);
    }
    /**
     * Sets the Connection
     *
     * @param connection The Solana Connection
     */
    async setConnection(connection) {
        this.connection = connection;
    }
    async fetchTokenDecimals(tokenAddr, chain) {
        if (!this.connection)
            throw new Error('no connection');
        let mint = await this.connection.getParsedAccountInfo(new web3_js_1.PublicKey(tokenAddr));
        if (!mint)
            throw new Error('could not fetch token details');
        const { decimals } = mint.value.data.parsed.info;
        return decimals;
    }
    /**
     * Gets the owner address of an Associated Token Account
     *
     * @param accountAddr The associated token account address
     * @returns The owner address
     */
    async getTokenAccountOwner(accountAddr) {
        const token = await (0, spl_token_1.getAccount)(this.connection, new web3_js_1.PublicKey(accountAddr));
        return token.owner.toString();
    }
    async getNativeBalance(walletAddress, chain) {
        if (!this.connection)
            throw new Error('no connection');
        const balance = await this.connection.getBalance(new web3_js_1.PublicKey(walletAddress));
        return ethers_1.BigNumber.from(balance);
    }
    async getTokenBalance(walletAddress, tokenId, chain) {
        if (!this.connection)
            throw new Error('no connection');
        const address = await this.getForeignAsset(tokenId, chain);
        if (!address)
            return null;
        const splToken = await this.connection.getTokenAccountsByOwner(new web3_js_1.PublicKey(walletAddress), { mint: new web3_js_1.PublicKey(address) });
        if (!splToken.value[0])
            return null;
        const balance = await this.connection.getTokenAccountBalance(splToken.value[0].pubkey);
        return ethers_1.BigNumber.from(balance.value.amount);
    }
    /**
     * Gets the Associate Token Address
     *
     * @param token The token id (home chain/address)
     * @param account The wallet address
     * @returns The associated token address
     */
    async getAssociatedTokenAddress(token, account) {
        const solAddr = await this.mustGetForeignAsset(token, SOLANA_CHAIN_NAME);
        return await (0, spl_token_1.getAssociatedTokenAddress)(new web3_js_1.PublicKey(solAddr), new web3_js_1.PublicKey(account), undefined, spl_token_1.TOKEN_PROGRAM_ID, spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID);
    }
    /**
     * Gets the Associate Token Account
     *
     * @param token The token id (home chain/address)
     * @param account The wallet address
     * @returns The account, or null if it does not exist
     */
    async getAssociatedTokenAccount(token, account) {
        if (!this.connection)
            throw new Error('no connection');
        const addr = await this.getAssociatedTokenAddress(token, account);
        try {
            const account = await (0, spl_token_1.getAccount)(this.connection, addr);
            return account;
        }
        catch (_) {
            return null;
        }
    }
    /**
     * Creates the Associated Token Account for a given wallet address. This must exist before a user can send a token bridge transfer, also it is the recipient address when sending the transfer.
     *
     * @param token The token id (home chain/address)
     * @param account The wallet address
     * @param commitment The commitment level
     * @returns The transaction for creating the Associated Token Account
     */
    async createAssociatedTokenAccount(token, account, commitment) {
        if (!this.connection)
            throw new Error('no connection');
        const tokenAccount = await this.getAssociatedTokenAccount(token, account);
        if (tokenAccount)
            return;
        const solAddr = await this.mustGetForeignAsset(token, SOLANA_CHAIN_NAME);
        const associatedAddr = await this.getAssociatedTokenAddress(token, account);
        const payerPublicKey = new web3_js_1.PublicKey(account);
        const tokenPublicKey = new web3_js_1.PublicKey(solAddr);
        const associatedPublicKey = new web3_js_1.PublicKey(associatedAddr);
        const createAccountInst = (0, spl_token_1.createAssociatedTokenAccountInstruction)(payerPublicKey, associatedPublicKey, payerPublicKey, tokenPublicKey);
        const transaction = new web3_js_1.Transaction().add(createAccountInst);
        const { blockhash } = await this.connection.getLatestBlockhash(commitment);
        transaction.recentBlockhash = blockhash;
        transaction.feePayer = payerPublicKey;
        return transaction;
    }
    /**
     * Prepare the transfer instructions for a native token bridge transfer from Solana
     *
     * @dev This _must_ be claimed on the destination chain, see {@link Wormhole#completeTransfer | completeTransfer}
     *
     * @param senderAddress The address of the sender
     * @param amount The token amount to be sent
     * @param recipientChain The destination chain name or id
     * @param recipientAddress The associated token address where funds will be sent
     * @param relayerFee The fee that would be paid to a relayer
     * @param payload Arbitrary bytes that can contain any addition information about a given transfer
     * @param commitment The commitment level
     * @returns The transaction for sending Native SOL from Solana
     */
    async transferNativeSol(senderAddress, amount, recipientChain, recipientAddress, relayerFee, payload, commitment) {
        if (!this.connection)
            throw new Error('no connection');
        const contracts = this.contracts.mustGetContracts(SOLANA_CHAIN_NAME);
        if (!contracts.core || !contracts.token_bridge) {
            throw new Error('contracts not found');
        }
        const rentBalance = await (0, spl_token_1.getMinimumBalanceForRentExemptAccount)(this.connection, commitment);
        const payerPublicKey = new web3_js_1.PublicKey(senderAddress);
        const ancillaryKeypair = web3_js_1.Keypair.generate();
        //This will create a temporary account where the wSOL will be created.
        const createAncillaryAccountIx = web3_js_1.SystemProgram.createAccount({
            fromPubkey: payerPublicKey,
            newAccountPubkey: ancillaryKeypair.publicKey,
            lamports: rentBalance,
            space: spl_token_1.ACCOUNT_SIZE,
            programId: spl_token_1.TOKEN_PROGRAM_ID,
        });
        //Send in the amount of SOL which we want converted to wSOL
        const initialBalanceTransferIx = web3_js_1.SystemProgram.transfer({
            fromPubkey: payerPublicKey,
            lamports: amount,
            toPubkey: ancillaryKeypair.publicKey,
        });
        //Initialize the account as a WSOL account, with the original payerAddress as owner
        const initAccountIx = (0, spl_token_1.createInitializeAccountInstruction)(ancillaryKeypair.publicKey, spl_token_1.NATIVE_MINT, payerPublicKey);
        //Normal approve & transfer instructions, except that the wSOL is sent from the ancillary account.
        const approvalIx = (0, tokenBridge_1.createApproveAuthoritySignerInstruction)(contracts.token_bridge, ancillaryKeypair.publicKey, payerPublicKey, amount);
        const message = web3_js_1.Keypair.generate();
        const nonce = (0, connect_sdk_1.createNonce)().readUInt32LE(0);
        const tokenBridgeTransferIx = payload
            ? (0, tokenBridge_1.createTransferNativeWithPayloadInstruction)(this.connection, contracts.token_bridge, contracts.core, senderAddress, message.publicKey, ancillaryKeypair.publicKey, spl_token_1.NATIVE_MINT, nonce, amount, Buffer.from(recipientAddress), this.wormhole.toChainId(recipientChain), payload)
            : (0, tokenBridge_1.createTransferNativeInstruction)(this.connection, contracts.token_bridge, contracts.core, senderAddress, message.publicKey, ancillaryKeypair.publicKey, spl_token_1.NATIVE_MINT, nonce, amount, relayerFee || BigInt(0), Buffer.from(recipientAddress), this.wormhole.toChainId(recipientChain));
        //Close the ancillary account for cleanup. Payer address receives any remaining funds
        const closeAccountIx = (0, spl_token_1.createCloseAccountInstruction)(ancillaryKeypair.publicKey, //account to close
        payerPublicKey, //Remaining funds destination
        payerPublicKey);
        const { blockhash } = await this.connection.getLatestBlockhash(commitment);
        const transaction = new web3_js_1.Transaction();
        transaction.recentBlockhash = blockhash;
        transaction.feePayer = payerPublicKey;
        transaction.add(createAncillaryAccountIx, initialBalanceTransferIx, initAccountIx, approvalIx, tokenBridgeTransferIx, closeAccountIx);
        transaction.partialSign(message, ancillaryKeypair);
        return transaction;
    }
    /**
     * Prepare the transfer instructions for a token bridge transfer from Solana
     *
     * @dev This _must_ be claimed on the destination chain, see {@link Wormhole#completeTransfer | completeTransfer}
     *
     * @param senderAddress The address of the sender
     * @param amount The token amount to be sent
     * @param recipientChain The destination chain name or id
     * @param recipientAddress The associated token address where funds will be sent
     * @param fromAddress The token account pubkey, owned by fromOwner address
     * @param tokenChainId The id of the token's chain
     * @param mintAddress The token address on the destination
     * @param fromOwnerAddress If not specified, will default to the sender address
     * @param relayerFee The fee that would be paid to a relayer
     * @param payload Arbitrary bytes that can contain any addition information about a given transfer
     * @param commitment The commitment level
     * @returns The transaction for sending tokens from Solana
     */
    async transferFromSolana(senderAddress, amount, recipientChain, recipientAddress, fromAddress, // token account pubkey, owned by fromOwner address
    tokenChainId, mintAddress, // token address
    fromOwnerAddress, relayerFee, payload, commitment) {
        if (!this.connection)
            throw new Error('no connection');
        const contracts = this.contracts.mustGetContracts(SOLANA_CHAIN_NAME);
        if (!contracts.core || !contracts.token_bridge) {
            throw new Error('contracts not found');
        }
        const recipientChainId = this.wormhole.toChainId(recipientChain);
        if (fromOwnerAddress === undefined) {
            fromOwnerAddress = senderAddress;
        }
        const nonce = (0, connect_sdk_1.createNonce)().readUInt32LE(0);
        const approvalIx = (0, tokenBridge_1.createApproveAuthoritySignerInstruction)(contracts.token_bridge, fromAddress, new web3_js_1.PublicKey(fromOwnerAddress), amount);
        const message = web3_js_1.Keypair.generate();
        const tokenBridgeTransferIx = payload
            ? (0, tokenBridge_1.createTransferWrappedWithPayloadInstruction)(this.connection, contracts.token_bridge, contracts.core, senderAddress, message.publicKey, fromAddress, fromOwnerAddress, tokenChainId, mintAddress, nonce, amount, recipientAddress, recipientChainId, payload)
            : (0, tokenBridge_1.createTransferWrappedInstruction)(this.connection, contracts.token_bridge, contracts.core, senderAddress, message.publicKey, fromAddress, fromOwnerAddress, tokenChainId, mintAddress, nonce, amount, relayerFee || BigInt(0), recipientAddress, recipientChainId);
        const transaction = new web3_js_1.Transaction().add(approvalIx, tokenBridgeTransferIx);
        const { blockhash } = await this.connection.getLatestBlockhash(commitment);
        transaction.recentBlockhash = blockhash;
        transaction.feePayer = new web3_js_1.PublicKey(senderAddress);
        transaction.partialSign(message);
        return transaction;
    }
    async startTransfer(token, amount, sendingChain, senderAddress, recipientChain, recipientAddress, relayerFee, commitment) {
        if (!this.connection)
            throw new Error('no connection');
        const destContext = this.wormhole.getContext(recipientChain);
        const formattedRecipient = (0, utils_1.arrayify)(destContext.formatAddress(recipientAddress));
        const relayerFeeBN = relayerFee ? BigInt(relayerFee) : undefined;
        if (token === connect_sdk_1.NATIVE) {
            return await this.transferNativeSol(senderAddress, amount, recipientChain, formattedRecipient, relayerFeeBN, undefined, 'finalized');
        }
        else {
            const tokenContext = this.wormhole.getContext(token.chain);
            const formattedTokenAddr = (0, utils_1.arrayify)(await tokenContext.formatAssetAddress(token.address));
            const solTokenAddr = await this.mustGetForeignAsset(token, SOLANA_CHAIN_NAME);
            const splToken = await this.connection.getTokenAccountsByOwner(new web3_js_1.PublicKey(senderAddress), { mint: new web3_js_1.PublicKey(solTokenAddr) });
            if (!splToken || !splToken.value[0])
                throw new Error('account does not have any token balance');
            return await this.transferFromSolana(senderAddress, amount, recipientChain, formattedRecipient, splToken.value[0].pubkey, this.wormhole.toChainId(token.chain), formattedTokenAddr, undefined, relayerFeeBN, undefined, 'finalized');
        }
    }
    async startTransferWithPayload(token, amount, sendingChain, senderAddress, recipientChain, recipientAddress, payload, commitment) {
        if (!this.connection)
            throw new Error('no connection');
        const destContext = this.wormhole.getContext(recipientChain);
        const formattedRecipient = (0, utils_1.arrayify)(destContext.formatAddress(recipientAddress));
        if (token === connect_sdk_1.NATIVE) {
            return await this.transferNativeSol(senderAddress, amount, recipientChain, formattedRecipient, undefined, payload, 'finalized');
        }
        else {
            const tokenContext = this.wormhole.getContext(token.chain);
            const formattedTokenAddr = (0, utils_1.arrayify)(await tokenContext.formatAssetAddress(token.address));
            const solTokenAddr = await this.mustGetForeignAsset(token, SOLANA_CHAIN_NAME);
            const splToken = await this.connection.getTokenAccountsByOwner(new web3_js_1.PublicKey(senderAddress), { mint: new web3_js_1.PublicKey(solTokenAddr) });
            if (!splToken || !splToken.value[0])
                throw new Error('account does not have any token balance');
            return await this.transferFromSolana(senderAddress, amount, recipientChain, formattedRecipient, splToken.value[0].pubkey, this.wormhole.resolveDomain(token.chain), formattedTokenAddr, undefined, undefined, payload, 'finalized');
        }
    }
    formatAddress(address) {
        const addr = typeof address === 'string' && address.startsWith('0x')
            ? (0, utils_1.arrayify)(address)
            : address;
        return (0, utils_1.arrayify)((0, utils_1.zeroPad)(new web3_js_1.PublicKey(addr).toBytes(), 32));
    }
    parseAddress(address) {
        const addr = typeof address === 'string' && address.startsWith('0x')
            ? (0, utils_1.arrayify)(address)
            : address;
        return new web3_js_1.PublicKey(addr).toString();
    }
    async formatAssetAddress(address) {
        return this.formatAddress(address);
    }
    async parseAssetAddress(address) {
        return this.parseAddress(address);
    }
    async getForeignAsset(tokenId, chain) {
        if (!this.connection)
            throw new Error('no connection');
        const chainId = this.wormhole.toChainId(tokenId.chain);
        const toChainId = this.wormhole.toChainId(chain);
        if (toChainId === chainId)
            return tokenId.address;
        const contracts = this.wormhole.mustGetContracts(chain);
        if (!contracts.token_bridge)
            throw new Error('contracts not found');
        const tokenContext = this.wormhole.getContext(tokenId.chain);
        const formattedAddr = await tokenContext.formatAssetAddress(tokenId.address);
        const addr = await (0, utils_2.getForeignAssetSolana)(this.connection, contracts.token_bridge, chainId, (0, utils_1.arrayify)(formattedAddr));
        if (!addr)
            return null;
        return addr;
    }
    async mustGetForeignAsset(tokenId, chain) {
        const addr = await this.getForeignAsset(tokenId, chain);
        if (!addr)
            throw new Error('token not registered');
        return addr;
    }
    async parseMessageFromTx(tx, chain) {
        if (!this.connection)
            throw new Error('no connection');
        const contracts = this.contracts.mustGetContracts(SOLANA_CHAIN_NAME);
        if (!contracts.core || !contracts.token_bridge)
            throw new Error('contracts not found');
        const response = await this.connection.getTransaction(tx);
        const parsedResponse = await this.connection.getParsedTransaction(tx);
        if (!response || !response.meta?.innerInstructions[0].instructions)
            throw new Error('transaction not found');
        const instructions = response.meta?.innerInstructions[0].instructions;
        const accounts = response.transaction.message.accountKeys;
        // find the instruction where the programId equals the Wormhole ProgramId and the emitter equals the Token Bridge
        const bridgeInstructions = instructions.filter((i) => {
            const programId = accounts[i.programIdIndex].toString();
            const emitterId = accounts[i.accounts[2]];
            const wormholeCore = contracts.core;
            const tokenBridge = (0, wormhole_1.deriveWormholeEmitterKey)(contracts.token_bridge);
            return programId === wormholeCore && emitterId.equals(tokenBridge);
        });
        const { message } = await (0, wormhole_1.getPostedMessage)(this.connection, accounts[bridgeInstructions[0].accounts[1]]);
        const parsedInstr = parsedResponse?.meta?.innerInstructions[0].instructions;
        const gasFee = parsedInstr
            ? parsedInstr.reduce((acc, c) => {
                if (!c.parsed || !c.parsed.info || !c.parsed.info.lamports)
                    return acc;
                return acc + c.parsed.info.lamports;
            }, 0)
            : 0;
        // parse message payload
        const parsed = (0, connect_sdk_1.parseTokenTransferPayload)(message.payload);
        // get sequence
        const sequence = response.meta?.logMessages
            ?.filter((msg) => msg.startsWith(SOLANA_SEQ_LOG))?.[0]
            ?.replace(SOLANA_SEQ_LOG, '');
        if (!sequence) {
            throw new Error('sequence not found');
        }
        // format response
        const tokenContext = this.wormhole.getContext(parsed.tokenChain);
        const destContext = this.wormhole.getContext(parsed.toChain);
        const tokenAddress = await tokenContext.parseAssetAddress((0, utils_1.hexlify)(parsed.tokenAddress));
        const tokenChain = this.wormhole.toChainName(parsed.tokenChain);
        const toAddress = destContext.parseAddress((0, utils_1.hexlify)(parsed.to));
        const parsedMessage = {
            sendTx: tx,
            sender: accounts[0].toString(),
            amount: ethers_1.BigNumber.from(parsed.amount),
            payloadID: parsed.payloadType,
            recipient: toAddress,
            toChain: this.wormhole.toChainName(parsed.toChain),
            fromChain: this.wormhole.toChainName(chain),
            tokenAddress,
            tokenChain,
            tokenId: {
                chain: tokenChain,
                address: tokenAddress,
            },
            sequence: ethers_1.BigNumber.from(sequence),
            emitterAddress: this.wormhole.conf.network === connect_sdk_1.Network.MAINNET
                ? SOLANA_MAINNET_EMMITER_ID
                : SOLANA_TESTNET_EMITTER_ID,
            gasFee: ethers_1.BigNumber.from(gasFee),
            block: response.slot,
        };
        if (parsedMessage.payloadID === 3) {
            const destContext = this.wormhole.getContext(parsed.toChain);
            const parsedPayload = destContext.parseRelayerPayload(parsed.tokenTransferPayload);
            const parsedPayloadMessage = {
                ...parsedMessage,
                relayerPayloadId: parsedPayload.relayerPayloadId,
                recipient: destContext.parseAddress(parsedPayload.to),
                to: toAddress,
                relayerFee: parsedPayload.relayerFee,
                toNativeTokenAmount: parsedPayload.toNativeTokenAmount,
            };
            return [parsedPayloadMessage];
        }
        return [parsedMessage];
    }
    async completeTransfer(destChain, signedVAA, overrides, payerAddr) {
        if (!payerAddr)
            throw new Error('receiving wallet address required for completing transfer on Solana');
        if (!this.connection)
            throw new Error('no connection');
        const contracts = this.contracts.mustGetContracts(SOLANA_CHAIN_NAME);
        if (!contracts.core || !contracts.token_bridge) {
            throw new Error('contracts not found for solana');
        }
        const parsed = (0, connect_sdk_1.parseTokenTransferVaa)(signedVAA);
        const tokenChain = parsed.tokenChain;
        if (tokenChain === connect_sdk_1.MAINNET_CHAINS.solana) {
            return await (0, utils_2.redeemAndUnwrapOnSolana)(this.connection, contracts.core, contracts.token_bridge, payerAddr, signedVAA);
        }
        else {
            return await (0, utils_2.redeemOnSolana)(this.connection, contracts.core, contracts.token_bridge, payerAddr, signedVAA);
        }
    }
    async isTransferCompleted(destChain, signedVaa) {
        if (!this.connection)
            throw new Error('no connection');
        const parsed = (0, connect_sdk_1.parseVaa)((0, utils_1.arrayify)(signedVaa));
        const tokenBridge = this.contracts.mustGetBridge(destChain);
        return (0, wormhole_1.getClaim)(this.connection, tokenBridge.programId, parsed.emitterAddress, parsed.emitterChain, parsed.sequence, 'finalized').catch((e) => false);
    }
    async getCurrentBlock() {
        if (!this.connection)
            throw new Error('no connection');
        return await this.connection.getSlot();
    }
    parseRelayerPayload(payload) {
        throw new Error('relaying is not supported on solana');
    }
}
exports.SolanaContext = SolanaContext;
//# sourceMappingURL=context.js.map