"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SolContracts = void 0;
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const wormhole_1 = require("./utils/wormhole");
const tokenBridge_1 = require("./utils/tokenBridge");
const nftBridge_1 = require("./utils/nftBridge");
/**
 * @category Solana
 */
class SolContracts extends connect_sdk_1.ContractsAbstract {
    constructor(wormholeBase) {
        super();
        this.wormhole = wormholeBase;
        const tag = wormholeBase.network === connect_sdk_1.Network.MAINNET ? 'mainnet-beta' : 'devnet';
        this._contracts = new Map();
        const chains = (0, connect_sdk_1.filterByContext)(wormholeBase.conf, connect_sdk_1.Context.SOLANA);
        chains.forEach((c) => {
            this._contracts.set(c.key, c.contracts);
        });
    }
    get connection() {
        return this.wormhole.connection;
    }
    getContracts(chain) {
        const chainName = this.wormhole.toChainName(chain);
        return this._contracts.get(chainName);
    }
    mustGetContracts(chain) {
        const chainName = this.wormhole.toChainName(chain);
        const contracts = this._contracts.get(chainName);
        if (!contracts)
            throw new Error(`no Solana contracts found for ${chain}`);
        return contracts;
    }
    /**
     * Returns core wormhole contract for the chain
     *
     * @returns An interface for the core contract, undefined if not found
     */
    getCore(chain) {
        if (!this.connection)
            throw new Error('no connection');
        const contracts = this.wormhole.mustGetContracts('solana');
        if (!contracts.core)
            return;
        return (0, wormhole_1.createReadOnlyWormholeProgramInterface)(contracts.core, this.connection);
    }
    /**
     * Returns core wormhole contract for the chain
     *
     * @returns An interface for the core contract, errors if not found
     */
    mustGetCore(chain) {
        const core = this.getCore(chain);
        if (!core)
            throw new Error(`Core contract for domain ${chain} not found`);
        return core;
    }
    /**
     * Returns wormhole bridge contract for the chain
     *
     * @returns An interface for the bridge contract, undefined if not found
     */
    getBridge(chain) {
        if (!this.connection)
            throw new Error('no connection');
        const contracts = this.wormhole.mustGetContracts('solana');
        if (!contracts.token_bridge)
            return;
        return (0, tokenBridge_1.createReadOnlyTokenBridgeProgramInterface)(contracts.token_bridge, this.connection);
    }
    /**
     * Returns wormhole bridge contract for the chain
     *
     * @returns An interface for the bridge contract, errors if not found
     */
    mustGetBridge(chain) {
        const bridge = this.getBridge(chain);
        if (!bridge)
            throw new Error(`Bridge contract for domain ${chain} not found`);
        return bridge;
    }
    /**
     * Returns wormhole NFT bridge contract for the chain
     *
     * @returns An interface for the NFT bridge contract, undefined if not found
     */
    getNftBridge(chain) {
        if (!this.connection)
            throw new Error('no connection');
        const contracts = this.wormhole.mustGetContracts('solana');
        if (!contracts.nft_bridge)
            return;
        return (0, nftBridge_1.createReadOnlyNftBridgeProgramInterface)(contracts.nft_bridge, this.connection);
    }
    /**
     * Returns wormhole NFT bridge contract for the chain
     *
     * @returns An interface for the NFT bridge contract, errors if not found
     */
    mustGetNftBridge(chain) {
        const nftBridge = this.getNftBridge(chain);
        if (!nftBridge)
            throw new Error(`NFT Bridge contract for domain ${chain} not found`);
        return nftBridge;
    }
    /**
     * Returns wormhole Token Bridge Relayer contract for the chain
     *
     * @returns An interface for the Token Bridge Relayer contract, undefined if not found
     */
    getTokenBridgeRelayer(chain) {
        return undefined;
    }
    /**
     * Returns wormhole Token Bridge Relayer contract for the chain
     *
     * @returns An interface for the Token Bridge Relayer contract, errors if not found
     */
    mustGetTokenBridgeRelayer(chain) {
        throw new Error('relayer not deployed on Solana');
    }
}
exports.SolContracts = SolContracts;
//# sourceMappingURL=contracts.js.map